#!/usr/bin/env python3

import asyncio
import functools
import logging

import aiocells


logger = logging.getLogger()


async def async_main():

    clock = aiocells.ModClock()
    graph = aiocells.DependencyGraph()

    # Here, we simplify the previous demo by using a single variable to
    # store the time and a single printer to announce the modifications
    # when they happen. Because we are using 'compute_flow', the graph
    # is computed when any of the timers go off.
    time = aiocells.ModVariable(clock)
    printer = aiocells.ModPrinter(clock, time, "time changed to {value}")
    graph.add_precedence(time, printer)

    # Set the time after 1 second
    timer_1 = functools.partial(aiocells.timer, 1, time)
    graph.add_precedence(timer_1, time)

    # Set the time after 3 seconds
    timer_3 = functools.partial(aiocells.timer, 3, time)
    graph.add_precedence(timer_3, time)

    print()
    print("Demo will complete in 10 iterations. Ctrl-C to cancel.")
    print()

    iteration_count = 0
    while await aiocells.compute_flow(graph):
        iteration_count += 1
        if iteration_count > 10:
            await aiocells.cancel_flow(graph)


def main():
    asyncio.run(async_main())



import asyncio
import collections
import dataclasses
import inspect
import logging

import aiocells.aio as aio

REPEATER = "cells.flow.repeater"

logger = logging.getLogger(__name__)


def repeat(function):
    if not inspect.iscoroutinefunction(function):
        raise ValueError("Event source must be a coroutine "
                         f"function: {function=}")
    setattr(function, REPEATER, True)
    return function


def is_repeater(function):
    return getattr(function, REPEATER, False)


@dataclasses.dataclass
class FlowState:

    input_tasks: list


async def compute_flow(graph):

    logger.debug("enter")

    if not hasattr(graph, "__flow_state"):
        logger.debug("First invocation, initialising flow state")
        callables, input_tasks = aio.prepare_ready_set(graph.input_nodes)
        assert len(callables) == 0, \
               f"Input nodes must be coroutines: {callables}"
        graph.__flow_state = FlowState(input_tasks)

    flow_state = graph.__flow_state

    try:
        # This is inside 'try' block so that the 'finally' block is executed
        if len(flow_state.input_tasks) == 0:
            return len(flow_state.input_tasks)

        # Wait for at least one input node to complete
        logger.debug("Waiting for input tasks")
        completed_input_tasks, flow_state.input_tasks = await asyncio.wait(
            flow_state.input_tasks,
            return_when=asyncio.FIRST_COMPLETED
        )

        logger.debug("Input received")
        # Here, we create new tasks for any input functions whose current
        # task has completed. For example, for a socket reader task that
        # has read some data and returned, we want to create a new task so
        # that it can read some more data
        aio.raise_task_exceptions(completed_input_tasks)
        completed_input_functions = [
            task.aio_coroutine_function
            for task in completed_input_tasks
        ]
        callables, new_tasks = aio.prepare_ready_set(
            completed_input_functions
        )
        assert len(callables) == 0,\
               f"Input nodes must be coroutines: {callables}"
        flow_state.input_tasks |= new_tasks

        logger.debug("Computing dependent nodes")
        for node in graph.topological_ordering:
            if node in graph.input_nodes:
                continue
            logger.debug("Computing dependent node: %s", node)
            if inspect.iscoroutinefunction(node):
                await node()
            else:
                assert callable(node)
                node()

        return len(flow_state.input_tasks)
    except asyncio.CancelledError:
        await aio.cancel_tasks(flow_state.input_tasks)
        # Don't reraise - this is an expected error when cancelling the
        # flow with `cancel_flow`
    except Exception:
        await aio.cancel_tasks(flow_state.input_tasks)
        # Reraise - this is an unexpected exception
        raise
    finally:
        logger.debug("exit, len(flow_state.input_tasks)=%s",
                     len(flow_state.input_tasks))


async def cancel_flow(graph):
    if not hasattr(graph, "__flow_state"):
        raise Exception("Graph does not have __flow_state attribute")
    await aio.cancel_tasks(graph.__flow_state.input_tasks)



#!/usr/bin/env python3

import asyncio
import functools
import logging

import aiocells


logger = logging.getLogger()


async def async_main():

    clock = aiocells.ModClock()
    graph = aiocells.DependencyGraph()

    # Two completely unrelated sequences are added to the graph. They
    # run concurrently.

    time_1 = aiocells.ModVariable(clock)
    timer_1 = functools.partial(aiocells.timer, 1, time_1)
    printer_1 = aiocells.ModPrinter(clock, time_1, "time_1 changed to {value}")
    graph.add_precedence(timer_1, time_1)
    graph.add_precedence(time_1, printer_1)

    time_3 = aiocells.ModVariable(clock)
    timer_3 = functools.partial(aiocells.timer, 3, time_3)
    printer_3 = aiocells.ModPrinter(clock, time_3, "time_3 changed to {value}")
    graph.add_precedence(timer_3, time_3)
    graph.add_precedence(time_3, printer_3)

    # With a flow computation, when any of the input nodes returns, all
    # non-input nodes are computed in topological order.  When this happens, we
    # are generally only interested in nodes that change as a result of the
    # input node returning. So, in this case, we see a message from "time_1"
    # every second and a message from "time_3" every 3 seconds

    print()
    print("Demo will complete in 10 iterations. Ctrl-C to cancel.")
    print()

    iteration_count = 0
    while await aiocells.compute_flow(graph):
        iteration_count += 1
        if iteration_count > 10:
            await aiocells.cancel_flow(graph)


def main():
    asyncio.run(async_main())
